/*
 * Copyright (c) 2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Mar 11 2019
 */
#include "acmrsim.h"

#include <err.h>
#include <pthread.h>

#include <iostream>
#include <vector>

#include "boost/bind.hpp"
#include "gazebo/gazebo.hh"
#include "gazebo/physics/physics.hh"
#include "gazebo/common/common.hh"

#include "libmrsim.h"

using namespace gazebo;

/* compat macro to access renamed getters in gazebo >= 8 */
#if GAZEBO_MAJOR_VERSION < 8
# define GZGET(x)	Get ## x
#else
# define GZGET(x)	x
#endif

class mrsim_plugin : public ModelPlugin
{
  struct rotor {
    physics::JointPtr joint;
    physics::LinkPtr child, parent;
    double noise;
    double cf, ct;
    int spin;
  };

  pthread_t reader;
  pthread_mutex_t mrsim_mutex = PTHREAD_MUTEX_INITIALIZER;
  struct mrsim_s *mrsim;

  physics::LinkPtr base;
  std::vector<rotor> rotors;
  bool explicit_wrench;

  event::ConnectionPtr begin, end;

 public:

  /* --- Load -------------------------------------------------------------- */

  void
  Load(physics::ModelPtr model, sdf::ElementPtr sdf)
  {
    struct mrsim_param_s *params;
    std::string name, pty, spin;
    sdf::ElementPtr b, r, j;
    uint8_t nrotors;
    double cf, ct;
    double noise;
    rotor rotor;

    /* get base link */
    if (!sdf->HasElement("link")) {
      warnx("mrsim-gazebo needs a <link> tag");
      gzthrow("mrsim-gazebo needs a <link> tag");
    }
    name = sdf->Get<std::string>("link");
    base = model->GetLink(name);
    if (!base) {
      warnx("non-existent link base: '%s'", name.c_str());
      gzthrow("mrsim-gazebo non-existent link base");
    }


    /* get list of controlled rotors */
    if (sdf->HasElement("rotors")) {
      r = sdf->GetElement("rotors");

      cf = r->HasElement("cf") ? r->Get<double>("cf") : 6e-4;
      ct = r->HasElement("ct") ? r->Get<double>("ct") : 1e-5;
      noise = r->HasElement("noise") ? r->Get<double>("noise") : 0.;

      for(r = r->GetElement("rotor"); r; r = r->GetNextElement("rotor")) {
        /* joint */
        if (!r->HasElement("joint")) {
          warnx("mrsim-gazebo ignoring <rotor> without <joint> tag");
          continue;
        }
        j = r->GetElement("joint");

        name = j->GetValue()->GetAsString();
        rotor.joint = model->GetJoint(name);
        if (!rotor.joint) {
          warnx("mrsim-gazebo ignoring non-existent joint '%s'", name.c_str());
          continue;
        }
        if (rotor.joint->GetType() !=
            (physics::Joint::JOINT | physics::Joint::HINGE_JOINT)) {
          warnx("mrsim-gazebo ignoring joint '%s' not of the hinge type",
                name.c_str());
          continue;
        }
        rotor.child = rotor.joint->GetChild();
        rotor.parent = rotor.joint->GetParent();

        /* just for visual effect at startup */
#if GAZEBO_MAJOR_VERSION < 4
        rotor.joint->SetAngle(0, math::Rand::GetDblUniform(0, M_PI));
#else
        rotor.joint->SetPosition(0, ignition::math::Rand::DblUniform(0, M_PI));
#endif

        /* joint spin */
        if (j->HasAttribute("spin")) {
          spin = j->GetAttribute("spin")->GetAsString();
          if (spin == "cw") rotor.spin = -1;
          else if (spin == "ccw") rotor.spin = 1;
          else {
            warnx("mrsim-gazebo ignoring unknown spin '%s' (must be cw or ccw)",
                  spin.c_str());
            rotor.spin = 1;
          }
        } else
          rotor.spin = 1;

        /* cf, ct */
        rotor.cf = r->HasElement("cf") ? r->Get<double>("cf") : cf;
        rotor.ct = r->HasElement("ct") ? r->Get<double>("ct") : ct;

        /* axis noise */
        rotor.noise = r->HasElement("noise") ? r->Get<double>("noise") : noise;

        /* push */
        rotors.push_back(rotor);
      }
      nrotors = rotors.size();

    } else if (sdf->HasElement("allocation")) {
      /* get number of rotors from matrix */
      std::stringstream G(sdf->Get<std::string>("allocation"));
      double d;

      for(nrotors = 0; !G.fail(); nrotors++) G >> d;
      nrotors /= 6;
    }


    /* init mrsim */
    mrsim = mrsim_init(nrotors);
    if (!mrsim) gzthrow("mrsim-gazebo could not initialize libmrsim");

    common::Time snow = model->GetWorld()->GZGET(SimTime)();
    struct timespec now = { snow.sec, snow.nsec };
    mrsim_update_time(mrsim, now);

    if (sdf->HasElement("pty"))
      pty = sdf->Get<std::string>("pty");
    else
      pty = "/tmp/mrsim";

    if (mrsim_open_pty(mrsim, pty.c_str())) {
      warn("mrsim-gazebo cannot create pty");
      gzthrow("mrsim-gazebo cannot create pty");
    }

    params = mrsim_params(mrsim);
#if GAZEBO_MAJOR_VERSION < 7
    params->body.g =
      model->GetWorld()->GetPhysicsEngine()->GetGravity().GetLength();
#else
    params->body.g = model->GetWorld()->Gravity().Length();
#endif
    if (sdf->HasElement("rotors")) {
      r = sdf->GetElement("rotors");

      if (params->body.rotors > 0)
        params->motor.J = rotors[0].child->GetInertial()->GZGET(IZZ)();
      if (r->HasElement("J")) params->motor.J = r->Get<double>("J");
      if (r->HasElement("b")) params->motor.b = r->Get<double>("b");
      if (r->HasElement("K")) params->motor.K = r->Get<double>("K");
      if (r->HasElement("R")) params->motor.R = r->Get<double>("R");
      if (r->HasElement("L")) params->motor.L = r->Get<double>("L");
      if (r->HasElement("V")) params->motor.V = r->Get<double>("V");
    }

    explicit_wrench = sdf->HasElement("allocation");
    if (explicit_wrench) {
      std::stringstream G(sdf->Get<std::string>("allocation"));
      uint8_t i;

      for(i = 0; i < 6 * params->body.rotors; i++) {
        G >> params->body.G[i];
        if (G.fail()) {
          warnx("mrsim-gazebo ignoring bad allocation matrix");
          explicit_wrench = false;
          break;
        }
      }
      if (explicit_wrench) {
        G >> std::ws;
        if (!G.eof()) {
          warnx("mrsim-gazebo ignoring bad allocation matrix");
          explicit_wrench = false;
        }
      }
    }


    /* create reader thread */
    pthread_create(&reader, NULL, reader_thread, this);


    /* plug events */
    begin = event::Events::ConnectWorldUpdateBegin(
      boost::bind(&mrsim_plugin::update_begin, this, _1));
    end = event::Events::ConnectWorldUpdateEnd(
      boost::bind(&mrsim_plugin::update_end, this));

    warnx("mrsim created %s", pty.c_str());
  }

  void
  update_begin(const common::UpdateInfo &)
  {
    struct timespec now;
    std::vector<rotor>::const_iterator r;
    struct mrsimstate_wrench_s *wrench;
    struct mrsimstate_motor_s *motor;
#if GAZEBO_MAJOR_VERSION < 8
    math::Vector3 w2;
#else
    ignition::math::Vector3d w2;
#endif
    uint8_t i;

    /* update motors */
    clock_gettime(CLOCK_REALTIME, &now);
    mrsim_update_time(mrsim, now);
    mrsim_update_motors(mrsim);

    /* set rotor velocity for visual effect - scale down by 6.28 */
    for(i = 0; i < rotors.size(); i++) {
      motor = mrsim_motor(mrsim, i);
      rotors[i].joint->SetVelocity(0, rotors[i].spin * motor->w);
    }

    if (explicit_wrench) {
      mrsim_update_wrench(mrsim);
      wrench = mrsim_wrench(mrsim);

      base->AddRelativeForce(
#if GAZEBO_MAJOR_VERSION < 8
        math::Vector3
#else
        ignition::math::Vector3d
#endif
        (wrench->thrust[0], wrench->thrust[1], wrench->thrust[2]));
      base->AddRelativeTorque(
#if GAZEBO_MAJOR_VERSION < 8
        math::Vector3
#else
        ignition::math::Vector3d
#endif
        (wrench->torque[0], wrench->torque[1], wrench->torque[2]));
    } else {
      /* apply thrust to rotors and drag to base link */
      for(i = 0, r = rotors.begin(); r != rotors.end(); ++r, i++) {
        motor = mrsim_motor(mrsim, i);

        w2 = motor->w * motor->w * r->joint->GZGET(GlobalAxis)(0);
        if (r->noise > 0.) {
          w2 =
#if GAZEBO_MAJOR_VERSION < 8
            math::Quaternion(
              math::Rand::GetDblNormal(0., r->noise),
              math::Rand::GetDblNormal(0., r->noise),
              math::Rand::GetDblNormal(0., r->noise)).RotateVector(w2);
#else
            ignition::math::Quaterniond(
              ignition::math::Rand::DblNormal(0., r->noise),
              ignition::math::Rand::DblNormal(0., r->noise),
              ignition::math::Rand::DblNormal(0., r->noise)).RotateVector(w2);
#endif
        }

        r->child->AddForce(r->cf * w2);
        r->parent->AddTorque(- r->ct * rotors[i].spin * w2);
      }
    }
  }

  void
  update_end()
  {
#if GAZEBO_MAJOR_VERSION < 8
    math::Pose p;
    math::Vector3 v, w, a, dw;
#else
    ignition::math::Pose3d p;
    ignition::math::Vector3d v, w, a, dw;
#endif
    struct mrsimstate_body_s *state;

    /* get state and update mrsim */
    p = base->GZGET(WorldPose)();
    v = base->GZGET(WorldLinearVel)();
    w = base->GZGET(WorldAngularVel)();
    a = base->GZGET(WorldLinearAccel)();
    dw = base->GZGET(WorldAngularAccel)();

    state = mrsim_body(mrsim);
#if GAZEBO_MAJOR_VERSION < 8
    state->px = p.pos.x; state->py = p.pos.y; state->pz = p.pos.z;
    state->qw = p.rot.w;
    state->qx = p.rot.x; state->qy = p.rot.y; state->qz = p.rot.z;
    state->vx = v.x; state->vy = v.y; state->vz = v.z;
    state->wx = w.x; state->wy = w.y; state->wz = w.z;
    state->ax = a.x; state->ay = a.y; state->az = a.z;
    state->dwx = dw.x; state->dwy = dw.y; state->dwz = dw.z;
#else
    state->px = p.Pos().X(); state->py = p.Pos().Y(); state->pz = p.Pos().Z();
    state->qw = p.Rot().W();
    state->qx = p.Rot().X(); state->qy = p.Rot().Y(); state->qz = p.Rot().Z();
    state->vx = v.X(); state->vy = v.Y(); state->vz = v.Z();
    state->wx = w.X(); state->wy = w.Y(); state->wz = w.Z();
    state->ax = a.X(); state->ay = a.Y(); state->az = a.Z();
    state->dwx = dw.X(); state->dwy = dw.Y(); state->dwz = dw.Z();
#endif

    mrsim_update_imu(mrsim);

    /* send sensor messages */
    if (mrsim_send_msgs(mrsim))
      warn("mrsim_send_msgs");
  }

  ~mrsim_plugin()
  {
    struct mrsim_s *handle = mrsim;

    pthread_mutex_lock(&mrsim_mutex);
    mrsim = NULL;
    pthread_mutex_unlock(&mrsim_mutex);
    pthread_join(reader, NULL);
    mrsim_fini(handle);
  }


 private:
  static void *
  reader_thread(void *data)
  {
    mrsim_plugin *plugin = (mrsim_plugin *)data;
    int s;

    pthread_mutex_lock(&plugin->mrsim_mutex);
    while(plugin->mrsim) {
      pthread_mutex_unlock(&plugin->mrsim_mutex);

      s = mrsim_wait_msg(plugin->mrsim, 500);

      pthread_mutex_lock(&plugin->mrsim_mutex);
      if (!plugin->mrsim) break;
      if (s == 0) continue;

      if (s < 0 || mrsim_process_msgs(plugin->mrsim)) {
        warn("mrsim message processing");
        break;
      }
    }
    pthread_mutex_unlock(&plugin->mrsim_mutex);

    warnx("mrsim disconnecting");
    return NULL;
  }
};

// Register this plugin with the simulator
GZ_REGISTER_MODEL_PLUGIN(mrsim_plugin);
